import numpy as np
import pywavefront

from RenderCat import camera


class VBO:
    # noinspection PyDictCreation
    def __init__(self, ctx):
        self.vbos = {}
        self.vbos['cube'] = CubeVBO(ctx)
        self.vbos['skybox'] = SkyboxVBO(ctx)
        self.vbos['overlay'] = OverlayVBO(ctx)
        vbos_to_load = []
        try:
            with open(file='object_list.txt', mode='r') as vbolist:
                for line in vbolist:
                    value = line.split("|")
                    vbos_to_load.append(value)
                vbolist.close()
            vbolist.close()
        except PermissionError:
            raise Exception("Cannot open VBO list file: no permission.")
        except FileNotFoundError:
            raise Exception("VBO list not found!")
        for value in vbos_to_load:
            self.vbos[value[0]] = LoadVBO(ctx, file=f"Models/{value[1][:]}".rstrip())

    def destroy(self):
        [vbo.destroy() for vbo in self.vbos.values()]


class BaseVBO:
    def __init__(self, ctx):
        self.ctx = ctx
        self.vbo = self.get_vbo()
        self.format: str = None
        self.attribs: list = None

    def get_vertex_data(self): ...

    def get_vbo(self):
        vertex_data = self.get_vertex_data()
        vbo = self.ctx.buffer(vertex_data)
        return vbo

    def destroy(self):
        self.vbo.release()


class CubeVBO(BaseVBO):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.format = '2f 3f 3f'
        self.attribs = ['in_texcoord_0', 'in_normal', 'in_position']

    @staticmethod
    def get_data(vertices, indices):
        data = [vertices[ind] for triangle in indices for ind in triangle]
        return np.array(data, dtype='f4')

    def get_vertex_data(self):
        vertices = [(-1, -1, 1), (1, -1, 1), (1, 1, 1), (-1, 1, 1),
                    (-1, 1, -1), (-1, -1, -1), (1, -1, -1), (1, 1, -1)]

        indices = [(0, 2, 3), (0, 1, 2),
                   (1, 7, 2), (1, 6, 7),
                   (6, 5, 4), (4, 7, 6),
                   (3, 4, 5), (3, 5, 0),
                   (3, 7, 4), (3, 2, 7),
                   (0, 6, 1), (0, 5, 6)]
        vertex_data = self.get_data(vertices, indices)

        tex_coord_vertices = [(0, 0), (1, 0), (1, 1), (0, 1)]
        tex_coord_indices = [(0, 2, 3), (0, 1, 2),
                             (0, 2, 3), (0, 1, 2),
                             (0, 1, 2), (2, 3, 0),
                             (2, 3, 0), (2, 0, 1),
                             (0, 2, 3), (0, 1, 2),
                             (3, 1, 2), (3, 0, 1), ]
        tex_coord_data = self.get_data(tex_coord_vertices, tex_coord_indices)

        normals = [(0, 0, 1) * 6,
                   (1, 0, 0) * 6,
                   (0, 0, -1) * 6,
                   (-1, 0, 0) * 6,
                   (0, 1, 0) * 6,
                   (0, -1, 0) * 6, ]
        normals = np.array(normals, dtype='f4').reshape(36, 3)

        vertex_data = np.hstack([normals, vertex_data])
        vertex_data = np.hstack([tex_coord_data, vertex_data])
        return vertex_data


class SkyboxVBO(BaseVBO):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.format = '3f'
        self.attribs = ['in_position']

    def get_vertex_data(self):
        z = 0.99999
        vertices = [(-1, -1, z), (3, -1, z), (-1, 3, z)]
        vertex_data = np.array(vertices, dtype='f4')
        return vertex_data


class OverlayVBO(BaseVBO):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.format = '3f'
        self.attribs = ['in_position']

    def get_vertex_data(self):
        z = camera.ZNEAR
        vertices = [(-1, -1, z), (1, -1, z), (-1, 1, z),
                    (-1, 1, z), (1, -1, z), (1, 1, z)]
        vertex_data = np.array(vertices, dtype='f4')
        return vertex_data


class LoadVBO(BaseVBO):
    def __init__(self, app, file=None):
        self.filename = file
        super().__init__(app)
        self.format = '2f 3f 3f'
        self.attribs = ['in_texcoord_0', 'in_normal', 'in_position']

    # noinspection PyUnusedLocal
    def get_vertex_data(self):
        try:
            objs = pywavefront.Wavefront(file_name=self.filename, cache=True, parse=True)
        except PermissionError:
            raise Exception(f"Cannot open model file {self.filename}: no permission.")
        except FileNotFoundError:
            raise Exception("Model file not found!")
        obj = objs.materials.popitem()[1]
        vertex_data = obj.vertices
        vertex_data = np.array(vertex_data, dtype='f4')
        return vertex_data


if __name__ == '__main__':
    print("You are forbidden to be here...")
